# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Database class. Connects to MongoDB and abstracts all communication with it.
"""

import time
import uuid

from pymongo import MongoClient


class DatabaseV1(object):
    """
    Database class. Connects to MongoDB and abstracts all communication with it.
    """

    def __init__(self):
        self.client = MongoClient()
        self.db = self.client.moirai
        self.token_lifespan = 24 * 3600
        self.__create_indexes()

    def close(self):
        self.client.close()

    def set_setting(self, key, value):
        db = self.db.settings
        db.replace_one({"key": key}, {"key": key, "value": value}, upsert=True)

    def get_setting(self, key):
        db = self.db.settings
        document = db.find_one({"key": key})
        if document:
            return document['value']
        else:
            return None

    def verify_token(self, token):
        now = time.time()
        span = self.token_lifespan
        ts = self.get_setting('tokens')
        vs = [t for t in ts if t['token'] == token and now - t['time'] < span]
        if len(vs) > 0:
            ts = [t for t in ts if t['token'] != token]
            ts.append({'token': token, 'time': now})
            self.set_setting('tokens', ts)
            return True
        return False

    def generate_token(self):
        t = {
            'token': uuid.uuid4().hex,
            'time': time.time()
        }
        ts = self.get_setting('tokens') or []
        ts += [t]
        ts = [t for t in ts if time.time() - t['time'] < self.token_lifespan]
        self.set_setting('tokens', ts)
        return t['token']

    def save_test_sensor_value(self, test, sensor, value, time, start_time):
        db = self.db.test_sensor_values
        data = {
            'test': test,
            'sensor': sensor,
            'value': value,
            'time': time,
            'start_time': start_time
        }
        db.insert_one(data)

    def list_test_data(self):
        db = self.db.test_sensor_values
        cursor = db.aggregate([
            {
                '$match': {
                    'time': {
                        '$lt': 1
                    }
                }
            },
            {
                '$group': {
                    '_id': {
                        'name': '$test',
                        'date': '$start_time'
                    }
                }
            },
            {
                '$project': {
                    '_id': 0,
                    'name': '$_id.name',
                    'date': '$_id.date'
                }
            }
        ])

        return list(cursor)

    def get_test_data(self, test, start_time, skip=0):
        db = self.db.test_sensor_values

        cursor = db.aggregate([
            {
                '$match': {
                    'test': test,
                    'start_time': start_time
                }
            },
            {
                '$sort': {
                    'time': 1
                }
            },
            {
                '$skip': skip
            },
            {
                '$project': {
                    'sensor': 1,
                    'time': 1,
                    'value': 1,
                    '_id': 0
                }
            }
        ])

        return list(cursor)

    def get_filtered_test_data(self, test, start_time, sensors):
        db = self.db.test_sensor_values

        cursor = db.aggregate([
            {
                '$match': {
                    'test': test,
                    'start_time': start_time,
                    'sensor': {
                        '$in': sensors
                    }
                }
            },
            {
                '$sort': {
                    'time': 1
                }
            },
            {
                '$group': {
                    '_id': '$sensor',
                    'values': {
                        '$push': '$value'
                    },
                    'time': {
                        '$push': '$time'
                    }
                }
            },
            {
                '$project': {
                    'time': 1,
                    'sensor': '$_id',
                    'values': 1,
                    '_id': 0
                }
            }
        ])

        return list(cursor)

    def remove_test(self, test):
        if isinstance(test, list):
            self.db.test_sensor_values.delete_many({
                "$or": test
            })
        else:
            self.db.test_sensor_values.delete_many(test)

    def dump_database(self):
        test_sensor_values = self.db.test_sensor_values.find({}, {'_id': 0})
        settings = self.db.settings.find({}, {'_id': 0})
        return list(settings), list(test_sensor_values)

    def restore_database(self, settings, test_sensor_values):
        self.db.settings.drop()
        self.db.test_sensor_values.drop()
        self.db.settings.insert_many(settings)
        self.db.test_sensor_values.insert_many(test_sensor_values)

    def __create_indexes(self):
        self.db.test_sensor_values.create_index('time', name='time')
        self.db.test_sensor_values.create_index('st', name='start_time')
        self.db.test_sensor_values.create_index('test', name='test')



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import io
import zipfile
import hashlib
import json
import dateutil.parser
import logging
import os.path
import tempfile
import scipy.io as sio

from bson import json_util
from enum import Enum

from flask import Flask, request, send_file
from moirai.database import DatabaseV1
from moirai.hardware import Hardware


class APIv1:
    """
    Starts a WebServer for the API endpoint.
    """

    def __init__(self, processHandler):
        self.app = Flask(__name__)
        self.database = DatabaseV1()
        self.hardware = Hardware()
        self.ph = processHandler
        logging.getLogger('werkzeug').setLevel(logging.ERROR)

    def run(self):
        """
        Entry point for the class. Adds routes and starts listening.
        """
        @self.app.after_request
        def add_header(response):
            response.headers['Cache-Control'] = 'no-store'
            return response
        self.app.add_url_rule('/',
                              view_func=lambda: 'Moirai Control System\n')
        self.app.add_url_rule('/login',
                              view_func=self.login,
                              methods=['POST'])
        self.app.add_url_rule('/set-password',
                              view_func=self.set_password,
                              methods=['POST'])
        self.app.add_url_rule('/hardware/drivers',
                              view_func=self.hardware_drivers,
                              methods=['GET'])
        self.app.add_url_rule('/hardware/configuration',
                              view_func=self.hardware_set_configuration,
                              methods=['POST'])
        self.app.add_url_rule('/hardware/configuration',
                              view_func=self.hardware_get_configuration,
                              methods=['GET'])
        self.app.add_url_rule('/system_response/tests',
                              view_func=self.system_response_get_tests,
                              methods=['GET'])
        self.app.add_url_rule('/system_response/tests',
                              view_func=self.system_response_set_tests,
                              methods=['POST'])
        self.app.add_url_rule('/system_response/test/run',
                              view_func=self.system_response_run,
                              methods=['POST'])
        self.app.add_url_rule('/system_response/test/stop',
                              view_func=self.system_response_stop,
                              methods=['GET'])
        self.app.add_url_rule('/live_graph/tests',
                              view_func=self.live_graph_list_tests,
                              methods=['GET'])
        self.app.add_url_rule('/live_graph/test',
                              view_func=self.live_graph_get_test,
                              methods=['POST'])
        self.app.add_url_rule('/live_graph/test/remove',
                              view_func=self.live_graph_remove_test,
                              methods=['POST'])
        self.app.add_url_rule('/live_graph/test/export',
                              view_func=self.live_graph_export_mat,
                              methods=['POST'])
        self.app.add_url_rule('/controllers',
                              view_func=self.controller_set,
                              methods=['POST'])
        self.app.add_url_rule('/controllers',
                              view_func=self.controller_get,
                              methods=['GET'])
        self.app.add_url_rule('/controllers/run',
                              view_func=self.controller_run,
                              methods=['POST'])
        self.app.add_url_rule('/controllers/stop',
                              view_func=self.controller_stop,
                              methods=['GET'])
        self.app.add_url_rule('/dev/gen_dummy_tests',
                              view_func=self.dev_gen_dummy_tests,
                              methods=['GET'])
        self.app.add_url_rule('/db/dump',
                              view_func=self.dump_database,
                              methods=['GET'])
        self.app.add_url_rule('/db/restore',
                              view_func=self.restore_database,
                              methods=['POST'])
        self.app.run(host="0.0.0.0", port=5000, threaded=True)

    def verify_token(self):
        """
        Verifies the token sent as a HTTP Authorization header.
        """
        try:
            authorization = request.headers.get('Authorization')
            token = authorization.split(' ')[-1]
            return self.database.verify_token(token)
        except Exception as e:
            print(e)
            return False

    def login(self):
        """
        Authenticates the user.
        Should be called with a POST request containing the following body:

        {
            "password": string
        }

        @returns:
            On success, HTTP 200 Ok and body:

            {
                "token": string
            }

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        password = request.json.get('password', '')
        hasher = hashlib.sha512()
        hasher.update(bytes(password, 'utf-8'))
        password = hasher.hexdigest()
        saved_password = self.database.get_setting('password')
        if saved_password == password:
            return json.dumps({'token': self.database.generate_token()})
        return '{}', 403

    def set_password(self):
        """
        Sets the password. Required body:

        {
            "password": string
        }

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403
        password = request.json.get('password', '')
        hasher = hashlib.sha512()
        hasher.update(bytes(password, 'utf-8'))
        password = hasher.hexdigest()
        self.database.set_setting('password', password)
        return '{}'

    def hardware_drivers(self):
        """
        Returns a JSON object with all the available drivers, their setups, if
        any, and ports, if listable.

        @returns:
            {
                name: string,
                has_setup: boolean,
                setup_arguments: [
                    {
                        name: string,
                        default_value: string
                    }
                ],
                ports: [
                    {
                        id: number,
                        name: string,
                        analog: {
                            input: boolean,
                            output: boolean,
                            read_range: [number, number],
                            write_range: [number, number]
                        }
                        digital: {
                            input: boolean,
                            output: boolean,
                            pwm: boolean
                        }
                    }
                ]
            }
        """
        if not self.verify_token():
            return '{}', 403
        drivers = self.hardware.list_drivers()
        drivers = [{
            'name': driver,
            'has_setup': self.hardware.driver_has_setup(driver),
            'setup_arguments': self.hardware.driver_setup_arguments(driver),
            'ports': self.__ports_for_driver(driver)
        } for driver in drivers]
        return json.dumps(drivers)

    def hardware_set_configuration(self):
        """
        Saves the given driver configuration. It must be a POST request with
        the following body:

        {
            name: string,
            setup_arguments: [
                {
                    name: string,
                    value: string
                }
            ],
            ports: [
                {
                    id: number,
                    name: string | number,
                    alias: string,
                    type: number,
                    defaultValue: string
                }
            ],
            configurations: [
                {
                    port: number,
                    alias: string,
                    formula: string
                }
            ]
        }

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        configuration = request.json
        self.database.set_setting('hardware_configuration', configuration)
        return '{}'

    def hardware_get_configuration(self):
        """
        Returns the saved driver configuration. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            {
                name: string,
                setup_arguments: [
                    {
                        name: string,
                        value: string
                    }
                ],
                ports: [
                    {
                        id: number,
                        name: string | number,
                        alias: string,
                        type: number,
                        defaultValue: string
                    }
                ],
                configurations: [
                    {
                        port: number,
                        alias: string,
                        formula: string
                    }
                ]
            }

            or

            {} if there is no configuration saved

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        config = self.database.get_setting('hardware_configuration') or {}
        return json.dumps(config)

    def system_response_get_tests(self):
        """
        Returns the saved system response tests. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            [{
                id: number
                name: string
                type: string
                inputs: string[]
                output: string
                points: [{
                    x: number
                    y: number
                }]
                fixedOutputs: [{
                    alias: string
                    value: number
                }]
                logRate: number
            }]

            or

            [] if there is no configuration saved

            On failure, HTTP 403 Unauthorized and body:

            []
        """
        if not self.verify_token():
            return '{}', 403

        tests = self.database.get_setting('system_response_tests') or []
        return json.dumps(tests)

    def system_response_set_tests(self):
        """
        Sets the saved system response tests. It must be a POST request with
        the following body:

        [{
            id: number
            name: string
            type: string
            inputs: string[]
            output: string
            points: [{
                x: number
                y: number
            }]
            fixedOutputs: [{
                alias: string
                value: number
            }]
            logRate: number
        }]

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        tests = request.json
        self.database.set_setting('system_response_tests', tests)
        return '{}'

    def system_response_run(self):
        """
        Runs the given test. It must be a POST request with the following body:

        {
            test: number
        }

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        test = request.json['test']
        self.ph.send_command("hardware", "run_test", test)
        return '{}'

    def system_response_stop(self):
        """
        Stops the running test. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        self.database.set_setting('current_test', None)

        return '{}'

    def live_graph_list_tests(self):
        """
        Returns a list of available graphs. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            [
                {
                    name: string
                    date: string (ISO 8601)
                    running: boolean
                }
            ]

            On failure, HTTP 403 Unauthorized and body:

            []
        """
        if not self.verify_token():
            return '{}', 403
        running_test = self.database.get_setting('current_test')
        tests = self.database.list_test_data()

        if len(tests) == 0:
            return '[]'

        last_date = max([test['date'] for test in tests])
        tests = [{
            'name': t['name'],
            'date': t['date'].isoformat(),
            'running': t['name'] == running_test and t['date'] == last_date
        } for t in tests]

        return json.dumps(tests)

    def live_graph_get_test(self):
        """
        Returns a graph. It must be a POST request with following body:

        {
            test: string
            start_time: string (ISO 8601)
            skip?: number
        }

        @returns:
            On success, HTTP 200 Ok and body:

            [
                {
                    sensor: string
                    time: string | number
                    value: string | number
                }
            ]

            On failure, HTTP 403 Unauthorized and body:

            []
        """
        if not self.verify_token():
            return '{}', 403

        test = request.json['test']
        start_time = dateutil.parser.parse(request.json['start_time'])
        skip = request.json.get('skip', 0)

        points = self.database.get_test_data(test, start_time, skip)

        return json.dumps(points)

    def live_graph_remove_test(self):
        """
        Deletes a test. It must be a POST request with following body:

        {
            test: string
            start_time: string (ISO 8601)
        }

        or a list of elements like that.

        @returns:
            On success, HTTP 200 Ok and body:

            []

            On failure, HTTP 403 Unauthorized and body:

            []
        """
        if not self.verify_token():
            return '{}', 403

        ts = request.json
        if isinstance(ts, list):
            ts = [{
                'test': t['test'],
                'start_time': dateutil.parser.parse(t['start_time'])
            } for t in ts]
        else:
            ts = {
                'test': ts['test'],
                'start_time': dateutil.parser.parse(ts['start_time'])
            }

        self.database.remove_test(ts)

        return '[]'

    def live_graph_export_mat(self):
        """
        Generates a MATLAB's MAT file from data. It must be a POST request with
        following body:

        {
            test: string
            start_time: string (ISO 8601)
        }

        @returns:
            On success, HTTP 200 Ok and body:

            file-contents

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        test = request.json['test']
        start_time = dateutil.parser.parse(request.json['start_time'])
        v = request.json['variables']

        ks = list(v.keys())
        ds = self.database.get_filtered_test_data(test, start_time, ks)
        ts = ds[0]['time']
        ds = {d['sensor']: d['values'] for d in ds}
        ds['t'] = ts

        directory = tempfile.gettempdir()
        file_path = os.path.join(directory, 'test.mat')

        sio.savemat(file_path, ds)

        f = open(file_path, 'rb')

        return send_file(f, as_attachment=True, attachment_filename='data.mat')

    def controller_set(self):
        """
        Saves a controller. It must be a POST request with following body:

        [{
            id: number
            name: string
            tau: number
            runTime: number
            before: string
            controller: string
            after: string
            inputs: string[]
        }]

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        cs = request.json
        self.database.set_setting('controllers', cs)
        return '{}'

    def controller_get(self):
        """
        Get saved controllers. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            [{
                id: number
                name: string
                tau: number
                runTime: number
                before: string
                controller: string
                after: string
                inputs: string[]
            }]

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        cs = self.database.get_setting('controllers') or []
        return json.dumps(cs)

    def controller_run(self):
        """
        Runs the given controller. It must be a POST request with the following
        body:

        {
            controller: number
        }

        @returns: On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        controller = request.json['controller']
        self.ph.send_command("hardware", "run_controller", controller)
        return '{}'

    def controller_stop(self):
        """
        Stops the running controller. It must be a GET request.

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        self.database.set_setting('current_test', None)

        return '{}'

    def dump_database(self):
        """
        Dumps the database collections. Use for backup.

        @returns:
            On success, HTTP 200 Ok and body:

            {
                settings: []
                test_sensor_values: []
            }

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403

        settings, test_sensor_values = self.database.dump_database()
        data = {
            'settings': settings,
            'test_sensor_values': test_sensor_values
        }
        jsondata = json.dumps(data, default=json_util.default)

        del data, settings, test_sensor_values  # release memory

        zbuffer = io.BytesIO()
        with zipfile.ZipFile(zbuffer, "a", zipfile.ZIP_LZMA) as zip_file:
            zip_file.writestr('dump', jsondata)
        zbuffer.seek(0, 0)
        return send_file(zbuffer, as_attachment=True, attachment_filename='dump.zip', mimetype="application/octet-stream")

    def restore_database(self):
        """
        Restore the database collections. It must be a POST request with following body:

        [{
            settings: []
            test_sensor_values: []
        }]

        @returns:
            On success, HTTP 200 Ok and body:

            {}

            On failure, HTTP 403 Unauthorized and body:

            {}
        """
        if not self.verify_token():
            return '{}', 403
        db = self.database
        file = request.files['file']
        with zipfile.ZipFile(file, "r", zipfile.ZIP_LZMA) as zip_file:
            jsondata = zip_file.read('dump')
            d = json.loads(jsondata, object_hook=json_util.object_hook)
            del jsondata
            db.restore_database(d['settings'], d['test_sensor_values'])
        return '{}'

    def __ports_for_driver(self, driver):
        """
        Returns a list of encoded ports for the given driver.
        """
        ps = self.hardware.driver_ports(driver)
        ps = [self.__encode_port(p) for p in ps]
        return ps

    def __encode_port(self, port):
        """
        If the port is a Enum, use its value. If it's a value that can be dumped
        to JSON, use it, if not, raise an exception.
        """
        port_id = port['id']
        if isinstance(port['id'], Enum):
            port_id = port['id'].value
        else:
            try:
                port_id = json.dumps(port['id'])
            except:
                pass
        port['id'] = port_id
        return port

    def __decode_port(self, driver, port):
        """
        If the ID in port should be converted to Enum, do it. Else, keep it as
        it is.
        """
        if driver not in ahio.list_available_drivers():
            return port
        driver = ahio.new_driver(driver)
        ps = driver.available_pins()
        if len(ps) > 0:
            p = ps[0]
            if isinstance(p['id'], Enum):
                port['id'] = p['id'].__class__(port['id'])
        return port

    def dev_gen_dummy_tests(self):
        import datetime
        start_time = datetime.datetime.utcnow()
        ts = list(range(600))

        xs = ts
        ys = [x ** 2 for x in xs]
        zs = [2 * x for x in xs]

        vss = [xs, ys, zs]
        ns = ['x', 'y', 'z']

        for n, vs in zip(ns, vss):
            for x, t in zip(vs, ts):
                self.database.save_test_sensor_value(
                    'Dummy',
                    n,
                    x,
                    t,
                    start_time
                )

        return '[]'



#!/usr/bin/env python3
# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import glob
import os
import platform
import re
import shutil
import sys
import zipfile

from urllib.request import urlopen
from pathlib import Path

opt = os.path.join(os.path.abspath(os.sep), 'opt')


def pi_version():
    """Detect the version of the Raspberry Pi.  Returns either 1, 2 or
    None depending on if it's a Raspberry Pi 1 (model A, B, A+, B+),
    Raspberry Pi 2 (model B+), or not a Raspberry Pi.
    https://github.com/adafruit/Adafruit_Python_GPIO/blob/master/Adafruit_GPIO/Platform.py
    """
    if not os.path.isfile('/proc/cpuinfo'):
        return None
    # Check /proc/cpuinfo for the Hardware field value.
    # 2708 is pi 1
    # 2709 is pi 2
    # Anything else is not a pi.
    with open('/proc/cpuinfo', 'r') as infile:
        cpuinfo = infile.read()
    # Match a line like 'Hardware   : BCM2709'
    match = re.search('^Hardware\s+:\s+(\w+)$', cpuinfo,
                      flags=re.MULTILINE | re.IGNORECASE)
    if not match:
        # Couldn't find the hardware, assume it isn't a pi.
        return None
    if match.group(1) == 'BCM2708':
        # Pi 1
        return 1
    elif match.group(1) == 'BCM2709':
        # Pi 2
        return 2
    elif match.group(1) == 'BCM2835':
        # Pi 3
        return 3
    else:
        # Something else, not a pi.
        return None


class cd:
    """Context manager for changing the current working directory"""

    def __init__(self, newPath):
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def download_snap7_linux(use_sudo):
    sudo = 'sudo' if use_sudo else ''
    arch = {
        'x64': 'x86_64',
        'ia32': 'i386',
        'x86': 'i386',
        'armv7': 'arm_v7',
        'armv6': 'arm_v6'
    }.get(platform.machine(), platform.machine())
    print('Downloading snap7...')
    try:
        if not os.path.exists('snap7.zip'):
            snap7url = 'https://sourceforge.net/projects/snap7/files/1.4.2/snap7-full-1.4.2.7z/download'
            contents = urlopen(snap7url).read()
            with open('snap7.zip', 'wb') as f:
                f.write(contents)
        os.system('7z x -y snap7.zip')
        with cd('snap7-full-1.4.2/build/unix'):
            os.system(f'make -f {arch}_linux.mk')
        lib = f'snap7-full-1.4.2/build/bin/{arch}-linux/libsnap7.so'
        shutil.copy(lib, os.path.join(opt, 'libsnap7.so'))
        os.chmod(os.path.join(opt, 'libsnap7.so'), 0o777)
        os.system(f'{sudo} ldconfig /opt')
        os.remove('snap7.zip')
        shutil.rmtree('snap7-full-1.4.2')
        return True
    except:
        return False


def download_snap7_win():
    print('Downloading snap7...')
    try:
        snap7url = 'https://sourceforge.net/projects/snap7/files/latest/download?source=files'
        contents = urlopen(snap7url).read()
        with open('snap7.zip', 'wb') as f:
            f.write(contents)
        with zipfile.ZipFile('snap7.zip', 'r') as zip_ref:
            zip_ref.extractall('snap7')
        dll = list(glob.iglob('snap7/**/*.dll'))[0]
        shutil.copy(dll, os.path.join(opt, 'snap7.dll'))
        os.remove('snap7.zip')
        shutil.rmtree('snap7')
        return True
    except:
        return False


def install(use_sudo=False):
    try:
        osname = platform.system()
        if osname == 'Windows':
            print('You need to install MongoDB.')
            print('https://www.mongodb.com')
            print('')
        elif osname == 'Darwin':
            brew = os.popen('which brew').read()
            if len(brew) > 0:
                print('Using Homebrew to install dependencies.')
                print('Will now execute: [brew install mongodb snap7]')
                result = os.system('brew install mongodb snap7')
                if result == 0:
                    print('Installation finished.')
                else:
                    print('Something went wrong. Try installing manually:')
                    print('brew install mongodb snap7')
            else:
                print('Homebrew not installed, cannot continue alone.')
                print('You need to install MongoDB and snap7.')
                print('https://www.mongodb.com')
                print('https://sourceforge.net/projects/snap7/files')
        elif osname == 'Linux':
            success = False
            sudo = 'sudo ' if use_sudo else ''
            if pi_version() is not None:
                cmd = sudo + 'DEBIAN_FRONTEND=noninteractive apt-get install -y mysql-server build-essential p7zip-full'
                print('Using APT to install dependencies.')
                print(f'Will now execute: [{cmd}]')
                result = os.system(cmd)
                with open('%s/.moirai.json' % str(Path.home()), 'rw') as f:
                    c = f.read()
                    if not c:
                        c.write('{"database":{"adapter":"mysql"}}')
                if result == 0:
                    if download_snap7_linux(use_sudo):
                        print('Installation finished.')
                    else:
                        print('Something went wrong. Try installing manually:')
                        print('https://sourceforge.net/projects/snap7/files')
                else:
                    print('Something went wrong. Try installing manually:')
                    print(cmd)
                    print('After that, you need to install snap7.')
                    print('https://sourceforge.net/projects/snap7/files')
            elif len(os.popen('which zypper').read()) > 0:
                cmd1 = sudo + 'zypper -n in -t pattern devel_basis'
                cmd2 = sudo + 'zypper -n in mongodb p7zip'
                print('Using Zypper to install dependencies.')
                print(f'Will now execute: [{cmd1}]')
                result1 = os.system(cmd1)
                print(f'Will now execute: [{cmd2}]')
                result2 = os.system(cmd2)
                if result1 == 0 and result2 == 0:
                    if download_snap7_linux(use_sudo):
                        print('Installation finished.')
                    else:
                        print('Something went wrong. Try installing manually:')
                        print('https://sourceforge.net/projects/snap7/files')
                else:
                    print('Something went wrong. Try installing manually:')
                    print(cmd1)
                    print(cmd2)
                    print('After that, you need to install snap7.')
                    print('https://sourceforge.net/projects/snap7/files')
            elif len(os.popen('which apt-get').read()) > 0:
                cmd = sudo + 'apt-get install -y mongodb-server build-essential p7zip-full'
                print('Using APT to install dependencies.')
                print(f'Will now execute: [{cmd}]')
                result = os.system(cmd)
                if result == 0:
                    if download_snap7_linux(use_sudo):
                        print('Installation finished.')
                    else:
                        print('Something went wrong. Try installing manually:')
                        print('https://sourceforge.net/projects/snap7/files')
                else:
                    print('Something went wrong. Try installing manually:')
                    print(cmd)
                    print('After that, you need to install snap7.')
                    print('https://sourceforge.net/projects/snap7/files')
            elif len(os.popen('which dnf').read()) > 0:
                cmd = sudo + 'dnf install -y @development-tools p7zip mongodb-server'
                print('Using DNF to install dependencies.')
                print(f'Will now execute: [{cmd}]')
                result = os.system(cmd)
                if result == 0:
                    if download_snap7_linux(use_sudo):
                        print('Installation finished.')
                    else:
                        print('Something went wrong. Try installing manually:')
                        print('https://sourceforge.net/projects/snap7/files')
                else:
                    print('Something went wrong. Try installing manually:')
                    print(cmd)
                    print('After that, you need to install snap7.')
                    print('https://sourceforge.net/projects/snap7/files')
            elif len(os.popen('which yum').read()) > 0:
                cmd1 = sudo + 'yum groupinstall -y "Development Tools" "Development Libraries"'
                cmd2 = sudo + 'yum install -y p7zip mongodb-server'
                print('Using YUM to install dependencies.')
                print(f'Will now execute: [{cmd1}]')
                result1 = os.system(cmd1)
                print(f'Will now execute: [{cmd2}]')
                result2 = os.system(cmd2)
                if result1 == 0 and result2 == 0:
                    if download_snap7_linux(use_sudo):
                        print('Installation finished.')
                    else:
                        print('Something went wrong. Try installing manually:')
                        print('https://sourceforge.net/projects/snap7/files')
                else:
                    print('Something went wrong. Try installing manually:')
                    print(cmd1)
                    print(cmd2)
                    print('After that, you need to install snap7.')
                    print('https://sourceforge.net/projects/snap7/files')
            else:
                print('You need to install MongoDB and snap7.')
                print('https://www.mongodb.com')
                print('https://sourceforge.net/projects/snap7/files')
        elif osname == 'Windows':
            print('Trying to install MongoDB and snap7...')
            if download_snap7_win():
                print('Downloading MongoDB...')
                mongo = 'https://fastdl.mongodb.org/win32/mongodb-win32-x86_64-2008plus-ssl-3.4.9-signed.msi'
                contents = urlopen(mongo).read()
                with open('mongo.msi', 'wb') as f:
                    f.write(contents)
                result = os.system(
                    r'msiexec.exe /q /i mongo.msi INSTALLLOCATION="C:\Program Files\MongoDB\Server\3.4.9\" ADDLOCAL="all"')
                if result:
                    print('Installation finished.')
                else:
                    print('You need to install MongoDB.')
                    print('https://www.mongodb.com')
            else:
                print('You need to install MongoDB and snap7.')
                print('https://www.mongodb.com')
                print('https://sourceforge.net/projects/snap7/files')
        else:
            print('You need to install MongoDB and snap7.')
            print('https://www.mongodb.com')
            print('https://sourceforge.net/projects/snap7/files')
    except:
        print('Something went wrong. Try installing manually.')
        print('You need to install MongoDB and snap7.')
        print('https://www.mongodb.com')
        print('https://sourceforge.net/projects/snap7/files')



# -*- coding: utf-8; -*-
#
# Copyright (c) 2016 Álan Crístoffer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from multiprocessing import Pipe
from threading import Thread

from moirai.abstract_process_handler import AbstractProcessHandler
from moirai.decorators import decorate_all_methods, dont_raise
from moirai.webapi.api import APIv1
from moirai.webapi.cmd_processor import CommandProcessor


def main(pipe):
    """
    Entry point for this process.
    """
    handler = ProcessHandler(pipe)
    handler.run()


@decorate_all_methods(dont_raise)
class ProcessHandler(AbstractProcessHandler):
    """
    Manages this processes' lifecycle and handles IPC.
    """

    def __init__(self, pipe):
        self.cmd_processor = CommandProcessor(self)
        super().__init__('WebAPI', pipe)
        self.api = APIv1(self)
        self.thread = Thread(target=self.api.run, name='WebAPIThread')

    def quit(self):
        pass

    def process_command(self, sender, cmd, args):
        """
        Redirects the commands received to the CommandProcessor class.
        """
        if cmd:
            self.cmd_processor.process_command(sender, cmd, args)

    def loop(self):
        pass
